module mod_myfzero    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!         This modules provides the implementation          !!!!!!
    !!!!!!!                of a root search algorithm                 !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    implicit none

    contains

        double precision function myfzero(f,a0,b0)
        ! returns one root of f between a0 and b0 using a dichotomy algorithm
        ! f : function
        ! a0: minimal bound of the root
        ! b0: maximal bound of the root
            implicit none

            double precision, intent(in)  :: a0, b0
            double precision    :: a, b, tol
            interface
                double precision function f(x)
                    double precision,intent(in)::x
                end function f
            end interface

            if (a0>b0) then
                myfzero = a0
            else if ((f(a0) * f(b0)) > 0) then
                if (abs(f(a0))<abs(f(b0))) then
                    myfzero = a0
                else
                    myfzero = b0
                end if
            else
                tol = 1d-12
                a = a0
                b = b0
                do while (abs(a-b)>tol)
                    if (f(a)*f(0.5d0*(a+b))<0) then
                        b=0.5d0*(a+b)
                    else
                        a=0.5d0*(a+b)
                    end if
                end do
                myfzero = 0.5d0*(a+b)
            end if
        end function myfzero
end module mod_myfzero
