module mod_myinterp1

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!!  This modules provides a linear interpolation function    !!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    implicit none

    contains

        function myinterp1(ni, nj, xd, yd, x) result(res)
        ! returns the value 'res' of the interpolation in 'x'
        ! ni : number of data points 
        ! nj : dimension of y
        ! yd : matrix ni x nj of y data values 
        ! xd : vector of length nj of x data values 
        ! x  : vector of length nj in which the interpolation should be computed
        ! res: vector of length ni of interpolated values        
            implicit none

            integer, intent(in)             :: ni, nj
            double precision, intent(in)    :: x
            double precision, intent(in)    :: xd(ni)
            double precision, intent(in)    :: yd(ni,nj)
            double precision, dimension(nj) :: res
            
            double precision :: t
            integer i, j
            
            if (ni == 1) then
                do j=1,nj
                    res(j) = yd(1,j)
                end do
                return
            end if

            if (x<=xd(1)) then
                t = (x-xd(1))/(xd(2)-xd(1))
                do j = 1, nj
                    res(j) = (1d0 - t)*yd(1,j) + t*yd(2,j)
                    do while (res(j)*yd(1,j)<0)
                        t = t/1.1d0
                        res(j) = (1d0 - t)*yd(1,j) + t*yd(2,j)
                    end do
                end do
            else if (x>=xd(ni)) then
                t = (x-xd(ni))/(xd(ni-1)-xd(ni))
                do j = 1, nj
                    res(j) = t*yd(ni-1,j) + (1d0-t)*yd(ni,j)
                    do while (res(j)*yd(ni,j)<0)                       
                        t = t/1.1d0
                        res(j) = t*yd(ni-1,j) + (1d0-t)*yd(ni,j)
                    end do                    
                end do
            else
                do i = 2, ni
                    if ((xd(i-1)<=x).and.(x<xd(i))) then
                        t = (x - xd(i-1))/(xd(i) - xd(i-1))
                        do j = 1, nj
                            res(j) = (1d0 - t) * yd(i-1,j) + t * yd(i,j)
! ~                             if (res(j)*yd(i,j)<0) then 
! ~                                 print*, "top à la vachette"
! ~                                 print*, [x, xd(i-1), xd(i)]
! ~                                 print*, [t, yd(i-1,j), yd(i,j)]
! ~                             end if
                        end do
                        
                        exit
                    end if
                end do
            end if
        end function
end module mod_myinterp1
