module mod_params

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!            This modules declares variables                !!!!!!
    !!!!!!! variable values do not change throughout execution        !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    use mod_tools

    implicit none

    double precision, parameter                 :: scaling = 1d2
    integer, parameter                          :: nstates= 7
    ![yh, Pih] = tauchen(7, 0, 0.95, 0.001, 3)
    double precision, dimension(nstates,nstates), parameter :: Pih = &
          reshape((/  &
                                0.868834, 0.027332, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, &
                                0.131158, 0.872575, 0.039084, 0.000001, 0.000000, 0.000000, 0.000000, &
                                0.000008, 0.100089, 0.886145, 0.054657, 0.000002, 0.000000, 0.000000, &
                                0.000000, 0.000004, 0.074769, 0.890684, 0.074769, 0.000004, 0.000000, &
                                0.000000, 0.000000, 0.000002, 0.054657, 0.886145, 0.100089, 0.000008, &
                                0.000000, 0.000000, 0.000000, 0.000001, 0.039084, 0.872575, 0.131158, &
                                0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.027332, 0.868834 &
                                                /), shape(Pih))
    ! reshape starts filling the columns. Pih is actually the transpose of the above matrix
    double precision, dimension(nstates,nstates), parameter :: Pil = Pih
    double precision, dimension(nstates), parameter   :: zh = scaling*2d-1*exp((/ -0.009608, -0.006405, -0.003203, 0.0, &
                                                                    0.003203, 0.006405, 0.009608/))
    double precision, dimension(nstates), parameter   :: zl = zh/2d0
    double precision, parameter ::  beta    = 0.96d0
    double precision, parameter ::  R       = 1.02d0
    double precision, parameter ::  z1h     = scaling*0.05d0/1d1
    double precision, parameter ::  z1l     = scaling*0.01d0/1d1
    double precision, parameter ::  z0      = scaling*0.2d0

    double precision, parameter ::  sig     = 2.0d0 
    double precision, parameter ::  probh = 0.5d0

    double precision, parameter :: tol = 1d-10
    double precision, parameter :: eps = .5d-9

    integer, parameter          :: nk  = 201

    integer, parameter          :: len_asset_grid = 21

    public  :: beta, R, z1h, z1l, z0, zh, zl, Pih, Pil, sig, probh, tol, eps, len_asset_grid, nk

    contains
        double precision function u(c)!periodic utility function
            implicit none
            double precision, intent(in)    :: c
            double precision    :: cc
            if (c<0d0) then
                cc=1d-100
            else
                cc = c
            end if! ~
            u = (1d0-beta)*(cc**(1d0-sig) - 0d0)/(1d0-sig)
        end function u

        double precision function uinv(u)!inverse of u
            implicit none
            double precision, intent(in)    :: u
            uinv = (1d0 + u*(1d0-sig)/(1d0-beta))**(1d0/(1d0-sig))
        end function uinv

        double precision function up(c)!derivative of u
            implicit none
            double precision, intent(in)    :: c
            double precision    :: cc
            if (c<0d0) then
                cc=1d-100
            else
                cc = c
            end if
            up = (1d0-beta)*cc**(-sig)
        end function up

end module mod_params
