module mod_tools
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!! This modules provides the code for some useful functions  !!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    implicit none
    
    interface mylogspace
        module procedure mylogspaced, mylogspacei
    end interface
    
    contains 
        
        function mymap(n, f,xs) result(toR)
        ! applies f to the array xs of size n
        ! replicates the 'map' function of functional languages
            implicit none
            
            double precision  :: xs(n), toR(n)
            
            integer :: i, n
            interface
                double precision function f(x)
                    double precision,intent(in)::x
                end function f
            end interface
            
            do i=1,n
                xs(i) = f(xs(i))
            end do
            toR=xs
        end function mymap
        
        function mylinspace(n, a, b) result(toR)
        !returns an array with n elements, equally spaced between a and b (first element is a, last one is b). 
        ! if n==1, returns a. Bottom isf n<1.
            implicit none
            double precision, intent(in):: a, b
            integer, intent(in)         :: n
            double precision            :: toR(n)
            integer :: i
            if (n==1) then
                toR(1) = a
            else 
                do i=1,n
                    toR(i)=a+(b-a)*dble(i-1)/dble(n-1)
                end do
            end if
        end function mylinspace
        
        function mylogspaced(n, a, b) result(toR)
        !returns an array with n elements, from 10**a to 10**b
            implicit none
            double precision, intent(in):: a, b
            integer, intent(in)         :: n
            double precision            :: toR(n)
            integer :: i
            if (n==1) then
                toR(1) = 10d0**a
            else 
                do i=1,n
                    toR(i)=10d0**(a+(b-a)*dble(i-1)/dble(n-1))
                end do
            end if
        end function mylogspaced
        
        function mylogspacei(n, a, b) result(toR)
        !returns an array with n elements, from 10**a to 10**b
            implicit none
            integer, intent(in)         :: n, a, b
            double precision            :: toR(n)
            toR = mylogspaced(n, dble(a), dble(b))
        end function mylogspacei
        
        function mylogbspace(n, base, a, b) result(toR)
        !returns an array with n elements, from base**a to base**b
            implicit none
            double precision, intent(in):: a, b, base
            integer, intent(in)         :: n
            double precision            :: toR(n)
            integer :: i
            if (n==1) then
                toR(1) = base**a
            else 
                do i=1,n
                    toR(i)=base**(a+(b-a)*dble(i-1)/dble(n-1))
                end do
            end if
        end function mylogbspace
        
        double precision function myUnbox(xs)
        ! returns the first element of the array xs
            implicit none
            double precision, intent(in)    ::  xs(:)
            myUnbox = xs(1)
        end function myUnbox
        
        subroutine myprint(n, p, tab)
            implicit none
            integer, intent(in) :: n, p
            double precision, intent(in)    ::  tab(n, p)
            integer             :: i_n, i_p
            do i_n=1,n
                do i_p=1,p
                    write(*,"(100f20.4)",advance="no") tab(i_n,i_p)
                end do
                write(*,*) ""
            end do
        end subroutine myprint
        
        subroutine printresults(tottime, n_loc, kvv, s0vv,  bool)
        ! prints results on the terminal
            implicit none            
            double precision, intent(in)    ::  tottime, kvv(:), s0vv(:)
            integer, intent(in)             ::  n_loc
            logical, intent(in)             :: bool
            
            write(*,'(A)', advance='no'), "completed in "
            write(*,'(1000F10.2)', advance='no'), tottime
            write(*,'(A)'), " seconds."
            if (bool) then
                print*, "results:"
                print*, "kvv:"
                call myprint(1, n_loc, kvv)
                print*, "s0vv:"
                call myprint(1, n_loc, s0vv)
            end if
        end subroutine printresults
        
        subroutine savedata(outfile, kvv, s0vv)
        ! saves data in the file 'outfile'
            implicit none            
            double precision, intent(in)    ::  kvv(:), s0vv(:)
            character(len=*), intent(in)    ::  outfile
            integer                         ::  i_loc
            
            open(unit=20, file=outfile, ACTION="write", STATUS="replace")
            do i_loc=1,size(kvv)
                write(20, '(1000F50.20)', advance='no')  kvv(i_loc)
                write(20, '(1000F50.20)')  s0vv(i_loc)
            end do
            close(20)
        end subroutine savedata        
        
        function myiminpos(n, weights, tab) result(imin)
        ! returns the index of the smallest element in tab whose weight is > 0
            integer, intent(in)             ::  n
            double precision, intent(in)    ::  tab(n), weights(n)
            integer :: i, imin
            imin = 1
            do while (weights(imin)<1d-100)
                imin = imin+1
                if (imin > n) stop "no positive element in weigthts!"
            end do
            do i=(imin+1),n
                if ((tab(i)<tab(imin)).and.(weights(i)>1d-100)) imin=i
            end do
        end function myiminpos

        function mymatpow(n, tab, p) result(res)
        ! raises the matrix tab of size (n,n) to exponent p
            integer, intent(in)             ::  p, n
            double precision, intent(in)    ::  tab(n,n)
            double precision, dimension(n,n)::  res(n,n)
            integer ::  i

            res = 0d0
            do i=1,n
                res(i,i)=1d0
            end do

            do i=1, p
                res = matmul(tab,res)
            end do
        end function mymatpow
            
end module mod_tools
