module mod_vars

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!            This modules declares variables                !!!!!!
    !!!!!!! variable values are likely to change throughout execution !!!!!!
    !!!!!!! (commonly used in prec-savings-*-pfi)                     !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    use mod_params

    implicit none

    double precision, dimension(:), allocatable               :: kv, s0v
    double precision, dimension (size(Pih,2))                 :: sPih, sPil

    integer :: ik, ia, iz, nrep, c1, c2, cr
    double precision, dimension (len_asset_grid, size(zh))    :: Vh, Vl, newVh, newVl, gah, gal, oldgah, oldgal
    double precision, dimension (len_asset_grid)              :: asset_grid, V1h, V1l, g1ah, g1al
    double precision                                          :: probl, diffh, diffl, ahres, alres, k

    public  :: ik, ia, iz, nrep, c1, c2, cr, Vh, Vl, newVh, newVl, gah, gal, V1h, V1l, asset_grid
    public  :: g1ah, g1al, diffh, diffl, ahres, alres, k, sPih, sPil, kv, s0v

    contains

        ! global initialization routine (should only be called once at the beginning of the program)
        subroutine global_init()
            implicit none
            integer i_loc
            double precision, dimension (size(Pih,1), size(Pih,2))    :: tmpPih, tmpPil
            probl = 1d0 - probh
            allocate(kv(1:nk))
            allocate(s0v(1:nk))
            s0v  = 0d0

            !conditional distribution (as if we start in state 0 at the date 1)
            sPih = reshape(Pih((size(Pih,1)+1)/2,:), shape(sPih))
            sPil = reshape(Pil((size(Pil,1)+1)/2,:), shape(sPil))

            kv = mylogspace(nk, -1, 2)

            asset_grid = scaling*(/ (1d-1 * dble(i_loc), i_loc=0,len_asset_grid-1) /)
        end subroutine

        ! local initialization routine (should be called anytime the value of ik changes)
        subroutine local_init()
            integer :: ia_loc, iz_loc
            k       = kv(ik)
            nrep    = 0
            diffh   = 1d0
            diffl   = 1d0
            do iz_loc = 1, size(zh)
                Vh(:,iz_loc)   = u(zh(iz_loc))
                Vl(:,iz_loc)   = u(zl(iz_loc))
            end do
        end subroutine
end module mod_vars
