program prec_savings

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!!           Core program file for EZW preferences           !!!!!!
    !!!!!!!                                                           !!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    use mod_params
    use mod_vars
    use mod_myinterp1
    use mod_myfzero
    use mod_tools

    implicit none
    call system_clock(count_rate=cr)
    call system_clock(c1)
    call global_init() !global initialization
    kv = min(kv+sig-0.1d0+1d-8, kv(size(kv))) !shifts kv and imposes k>=sig or using EZ notations (below: the shifted k corresponds to the risk aversion parameter alpha).
    print*, "starting..."
    
    write(*, '(1000A15)', advance='no'), "k"
    write(*, '(1000A15)', advance='no'), "s0"
    write(*, '(1000A15)', advance='yes'), "#rep"
    do ik=1, size(kv)

        call local_init()! local initialization
        write(*, '(1000F15.8)', advance='no'), k

        do while((max(diffh, diffl)>tol).and.(nrep<10000)) !this loop computes the policy and value functions after date 2
            nrep = nrep + 1
            do ia = 1, size(asset_grid)
                do iz = 1, size(zh)
                    ahres = myfzero(fooh, eps, zh(iz)+R*asset_grid(ia)-eps);
                    newVh(ia,iz) = -fooVh(ahres);
                    gah(ia,iz) = ahres

                    alres = myfzero(fool, eps, zl(iz)+R*asset_grid(ia)-eps);
                    newVl(ia,iz) = -fooVl(alres);
                    gal(ia,iz) = alres
                end do
            end do
            diffh = maxval(abs(oldgah-gah))
            diffl = maxval(abs(oldgal-gal))
            Vh = newVh
            Vl = newVl
            oldgah = gah
            oldgal = gal
        end do
        
        do ia = 1, size(asset_grid) !this loop computes the policy and value functions at date 1
            ahres = myfzero(foo1h, eps, z1h+R*asset_grid(ia)-eps)
            g1ah(ia) = ahres
            V1h(ia) = -fooV1h(ahres)

            alres = myfzero(foo1l, eps, z1l+R*asset_grid(ia)-eps)
            g1al(ia) = alres
            V1l(ia) = -fooV1l(alres)
        end do
        s0v(ik)     = myfzero(foo0, eps, z0-eps) !computes the optimal savings at date 0
        write(*, '(1000F15.8)', advance='no'), s0v(ik)
        print'(1000i15)', nrep
    end do
    call system_clock(c2)

    call printresults(dble(c2-c1)/dble(cr), nk, kv, s0v, .false.) !displays results
    call savedata('res-ez.dat', kv, s0v) !saves results in 'res-ez.dat'

    contains

        double precision function expV(a, prob_vect, V)
            ! computes the expected continuation utility
            ! a         : value of savings
            ! prob_vect : vector of transition probabilities
            ! V         : matrix of value functions
            implicit none
            double precision, intent(in):: a
            double precision, dimension(nstates), intent(in) :: prob_vect
            double precision, dimension(len_asset_grid, nstates), intent(in) :: V
            double precision, dimension(nstates)  :: interpV
            double precision                      :: EexpV
            integer                               :: iz_loc, iz_max

            interpV = myinterp1(len_asset_grid, nstates, asset_grid, V, a)
            iz_max = myiminpos(nstates, prob_vect, -((1d0-sig)*interpV)**(1d0-k))

            EexpV  = 0d0
            do iz_loc = 1, nstates
                if (prob_vect(iz_loc) > 1d-100) then
                    EexpV = EexpV + prob_vect(iz_loc)*&
                      (interpV(iz_loc)/interpV(iz_max))**((1d0-k)/(1d0-sig))
                end if
            end do
            if (abs(EexpV) < 1d-100) then
                stop '0 in EexpV of expV is raised'
            end if
            expV   = beta * interpV(iz_max) * EexpV**((1d0-sig)/(1d0-k))
            if (isnan(expV)) then
              stop '"expV" is a NaN'
            end if
        end function expV

        double precision function fooVh(ap) !continuation utility after date 2 in state h
            implicit none
            double precision, intent(in) :: ap
            fooVh = -u(zh(iz) + R*asset_grid(ia) - ap) - expV(ap, Pih(iz,:), Vh);
            if (isnan(fooVh)) stop '"fooVh" is a NaN'
        end function fooVh

        double precision function fooVl(ap) !continuation utility after date 2 in state l
            implicit none
            double precision, intent(in) :: ap
            fooVl = -u(zl(iz) + R*asset_grid(ia) - ap) - expV(ap, Pil(iz,:), Vl);
            if (isnan(fooVl)) stop '"fooVl" is a NaN'
        end function fooVl
        
        double precision function fooV1h(ap) !continuation utility at date 1 in state h
            implicit none
            double precision, intent(in) :: ap
            fooV1h = -u(z1h + R*asset_grid(ia) - ap) - expV(ap, sPih, Vh);
            if (isnan(fooV1h)) stop '"fooV1h" is a NaN'
        end function fooV1h

        double precision function fooV1l(ap) !continuation utility at date 1 in state l
            implicit none
            double precision, intent(in) :: ap
            fooV1l = -u(z1l + R*asset_grid(ia) - ap) - expV(ap, sPil, Vl);
            if (isnan(fooV1l)) stop '"fooV1l" is a NaN'
        end function fooV1l

        double precision function foo(ap, prob_vect, V, ga, z, zval)
            ! computes the residual of the Euler equation
            ! ap        : savings
            ! prob_vect : vector of transition probabilities
            ! V         : matrix of value functions
            ! ga        : matrix of policy functions
            ! z         : vector of states
            ! zval      : current state
            implicit none
            double precision, intent(in)                                    ::  ap, zval
            double precision, dimension(nstates), intent(in)                ::  z, prob_vect
            double precision, dimension(len_asset_grid, nstates), intent(in)::  V, ga
            double precision, dimension(nstates)                            ::  interpV, interga
            integer                                                         ::  iz_loc, iz_max
            double precision                                                ::  EexpV, EcontU
            interpV = myinterp1(len_asset_grid, nstates, asset_grid, V, ap) 
            interga = myinterp1(len_asset_grid, nstates, asset_grid, ga, ap)

            iz_max = myiminpos(nstates, prob_vect, -((1d0-sig)*interpV)**(1d0-k))

            EexpV  = 0d0
            EcontU = 0d0
            do iz_loc = 1, nstates
                if (prob_vect(iz_loc)>1d-100) then
                    EexpV = EexpV + prob_vect(iz_loc) * &
                        (interpV(iz_loc)/interpV(iz_max))**((1d0-k)/(1d0-sig))
                    EcontU = EcontU + prob_vect(iz_loc)*up(z(iz_loc)+R*ap-interga(iz_loc))&
                    * (interpV(iz_loc)/interpV(iz_max))**((sig-k)/(1d0-sig))
                end if
            end do
            if (abs(EexpV) < 1d-100) then
                stop '0 in EexpV of foo is raised'
            end if

            foo = -up(zval + R*asset_grid(ia) - ap) + beta*R * EcontU * (EexpV**((k-sig)/(1d0-k)));
        end function foo
        
        double precision function fooh(ap) !residual of the Euler equation after date 2 in state h
            implicit none
            double precision, intent(in)    ::  ap
            fooh = foo(ap, Pih(iz,:), Vh, gah, zh, zh(iz))
            if (isnan(fooh)) stop '"fooh" is a NaN'
        end function fooh

        double precision function fool(ap) !residual of the Euler equation after date 2 in state l
            implicit none
            double precision, intent(in)    ::  ap
            fool = foo(ap, Pil(iz,:), Vl, gal, zl, zl(iz))
            if (isnan(fool)) stop '"fool" is a NaN'
        end function fool
        
        double precision function foo1h(ap) !residual of the Euler equation at date 1 in state h
            implicit none
            double precision, intent(in)    ::  ap
            foo1h = foo(ap, sPih, Vh, gah, zh, z1h)
            if (isnan(foo1h)) stop '"fooh" is a NaN'
        end function foo1h

        double precision function foo1l(ap) !residual of the Euler equation at date 1 in state l
            implicit none
            double precision, intent(in)    ::  ap
            foo1l = foo(ap, sPil, Vl, gal, zl, z1l)
            if (isnan(foo1l)) stop '"fool" is a NaN'
        end function foo1l

        double precision function foo0(ap) !residual of the Euler equation at date 0
            implicit none
            double precision, intent(in)    ::  ap
            double precision                ::  interpVh, intergah, interpVl, intergal
            double precision                ::  EexpV, EcontU
            interpVh = myUnbox(myinterp1(size(asset_grid), 1, asset_grid, V1h, ap))
            interpVl = myUnbox(myinterp1(size(asset_grid), 1, asset_grid, V1l, ap))
            intergah = myUnbox(myinterp1(size(asset_grid), 1, asset_grid, g1ah, ap))
            intergal = myUnbox(myinterp1(size(asset_grid), 1, asset_grid, g1al, ap))

            if ((interpVl/interpVh)**((1d0-k)/(1d0-sig))<1d0) then
! ~                 print*, "h"
                EexpV   = probh + probl * (interpVl/interpVh)**((1d0-k)/(1d0-sig))
                EcontU  = probh * up(z1h+R*ap-intergah) &
                    + probl * up(z1l+R*ap-intergal) * (interpVl/interpVh)**((sig-k)/(1d0-sig))
            else
! ~                 print*, "l"
                EexpV   = probl + probh * (interpVh/interpVl)**((1d0-k)/(1d0-sig))
                EcontU  = probl * up(z1l+R*ap-intergal) &
                    + probh * up(z1h+R*ap-intergah) * (interpVh/interpVl)**((sig-k)/(1d0-sig))
            end if
            foo0 = -up(z0 - ap) + beta*R * EcontU * (EexpV**((k-sig)/(1d0-k)));
            if (abs(EexpV) < 1d-100) then
                print*, [interpVh, interpVl, interpVh-interpVl]
                stop '0 in EexpV of foo0 is raised'
            end if

            if (isnan(foo0)) then
                print*, [EexpV, EcontU, foo0, ap, uinv(interpVh), uinv(interpVl)]
                stop '"foo0" is a NaN'
            end if

        end function foo0


end program prec_savings
